-- ToME - Tales of Maj'Eyal
-- Copyright (C) 2009, 2010, 2011 Nicolas Casalini
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Nicolas Casalini "DarkGod"
-- darkgod@te4.org


newTalent{
	name = "Recruiting",
	type = {"leadership/command", 1},
	require = lead_wil_req1,
	points = 5,
	mode = "passive",
	getMaxMercenaries = function(self, t)
		return (self:getTalentLevel(t) + 1) / 2
	end,
	info = function(self, t)
		local maxMercenaries = t.getMaxMercenaries(self, t)

		return ([[Use your charisma to hire new party members and keep the current members working together.
		You may have up to %d additional mercenaries in your party.
		Additionally, each level of this skill reduces the cost of hiring mercenaries.]]):
		format(maxMercenaries)
	end,
}

newTalent{
	name = "Orders",
	type = {"leadership/command", 2},
	require = lead_wil_req2,
	points = 5,
	mode = "passive",
	on_learn = function(self, t)
		for uid, e in pairs(game.level.entities) do
			if e.mercenary and e.summoner and e.summoner == self then
				game.party:removeMember(e)
				if self:getTalentLevelRaw(t) == 1 then
					game.party:addMember(e, {
						control="order",
						title="Mercenary",
						orders = {target=true},
					})
				elseif self:getTalentLevelRaw(t) == 2 then
					game.party:addMember(e, {
						control="order",
						title="Mercenary",
						orders = {target=true, leash=true},
					})
				elseif self:getTalentLevelRaw(t) == 3 then
					game.party:addMember(e, {
						control="order",
						title="Mercenary",
						orders = {target=true, leash=true, anchor=true},
					})
				elseif self:getTalentLevelRaw(t) == 4 then
					game.party:addMember(e, {
						control="order",
						title="Mercenary",
						orders = {target=true, leash=true, anchor=true, talents=true},
					})
				elseif self:getTalentLevelRaw(t) >= 5 then
					game.party:addMember(e, {
						control="order",
						title="Mercenary",
						orders = {target=true, leash=true, anchor=true, talents=true, behavior=true},
					})
				else
					game.party:addMember(e, {
						control="order",
						title="Mercenary",
					})
				end
			end
		end
	end,
	on_unlearn = function(self, t)
		for uid, e in pairs(game.level.entities) do
			if e.mercenary and e.summoner and e.summoner == self then
				game.party:removeMember(e)
				if self:getTalentLevelRaw(t) == 1 then
					game.party:addMember(e, {
						control="order",
						title="Mercenary",
						orders = {target=true},
					})
				elseif self:getTalentLevelRaw(t) == 2 then
					game.party:addMember(e, {
						control="order",
						title="Mercenary",
						orders = {target=true, leash=true},
					})
				elseif self:getTalentLevelRaw(t) == 3 then
					game.party:addMember(e, {
						control="order",
						title="Mercenary",
						orders = {target=true, leash=true, anchor=true},
					})
				elseif self:getTalentLevelRaw(t) == 4 then
					game.party:addMember(e, {
						control="order",
						title="Mercenary",
						orders = {target=true, leash=true, anchor=true, talents=true},
					})
				elseif self:getTalentLevelRaw(t) >= 5 then
					game.party:addMember(e, {
						control="order",
						title="Mercenary",
						orders = {target=true, leash=true, anchor=true, talents=true, behavior=true},
					})
				else
					game.party:addMember(e, {
						control="order",
						title="Mercenary",
					})
				end
			end
		end
	end,
	info = function(self, t)
		return ([[Each level of this skill opens up additional orders for your units.]])
	end,
}

newTalent{
	name = "Training",
	type = {"leadership/command", 3},
	require = lead_wil_req3,
	points = 5,
	mode = "passive",
	on_learn = function(self, t)
		for uid, e in pairs(game.level.entities) do
			if e.mercenary and e.summoner and e.summoner == self then
				e.inc_stats[e.STAT_STR] = e.inc_stats[e.STAT_STR] + 2
				e.inc_stats[e.STAT_DEX] = e.inc_stats[e.STAT_DEX] + 2
				e.inc_stats[e.STAT_CON] = e.inc_stats[e.STAT_CON] + 2
				e.inc_stats[e.STAT_MAG] = e.inc_stats[e.STAT_MAG] + 2
				e.inc_stats[e.STAT_WIL] = e.inc_stats[e.STAT_WIL] + 2
				e.inc_stats[e.STAT_CUN] = e.inc_stats[e.STAT_CUN] + 2
			end
		end
	end,
	on_unlearn = function(self, t)
		for uid, e in pairs(game.level.entities) do
			if e.mercenary and e.summoner and e.summoner == self then
				e.inc_stats[e.STAT_STR] = e.inc_stats[e.STAT_STR] - 2
				e.inc_stats[e.STAT_DEX] = e.inc_stats[e.STAT_DEX] - 2
				e.inc_stats[e.STAT_CON] = e.inc_stats[e.STAT_CON] - 2
				e.inc_stats[e.STAT_MAG] = e.inc_stats[e.STAT_MAG] - 2
				e.inc_stats[e.STAT_WIL] = e.inc_stats[e.STAT_WIL] - 2
				e.inc_stats[e.STAT_CUN] = e.inc_stats[e.STAT_CUN] - 2
			end
		end
	end,
	info = function(self, t)
		return ([[Rigorous training improves the stats of all of your hired mercenaries by %d.]]):format(self:getTalentLevel(t)*2)
	end,
}

newTalent{
	name = "Fast Talk",
	type = {"leadership/command", 4},
	require = lead_wil_req4,
	points = 5,
	cooldown = 100,
	no_npc_use = true,
	range = 4,
	requires_target = true,
	direct_hit = true,
	getPrice = function(self, t) return 60 - self:getTalentLevel(t) * 5 end,

	target = function(self, t) return {type="hit", range=self:getTalentRange(t), talent=t} end,
	on_pre_use = function(self, t, silent)
		if not self.mercs_in_party or self.mercs_in_party >= (self:getTalentLevel(self.T_RECRUITING) / 2 ) + 1 then
			if not silent then
				game.logPlayer(self, "You cannot handle another party member.")
			end
			return false
		elseif not (self.money > t.getPrice(self, t)) then
			if not silent then
				game.logPlayer(self, "You cannot afford the bribe.")
			end
			return false
		end
		return true
	end,

	action = function(self, t)
		local tg = self:getTalentTarget(t)
		local x, y = self:getTarget(tg)
		if not x or not y then return nil end
		self:project(tg, x, y, function(px, py)
			local target = game.level.map(px, py, Map.ACTOR)

			if not target or target.dead then return end

			if not target.type or not (target.type == "humanoid" or target.type == "giant" or (target.type == "undead" and self.undead)) then
				if self.undead then game.logSeen(target, "Only humanoids, giants, and undead can be hired.") return
				else game.logSeen(target, "Only humanoids and giants can be hired.") return end
			end

			if target.mercenary and target.summoner and target.summoner == self then
				game.logSeen(target, "%s already works for you.", target.name:capitalize())
				return
			end

			if not target:canBe("instakill") or target.no_hire or target.rank > 2 or not target:checkHit(self:combatTalentStatDamage(t, "wil", 15, 100) + self.level, target:getWil(20) + target.level) then
				game.logSeen(target, "%s refuses the bribe!", target.name:capitalize())
				return
			end

			self:incMoney(-t.getPrice(self, t))
			game.logSeen(target, "%s joins your party!", target.name:capitalize())
			target.faction = self.faction

			self.mercs_in_party = self.mercs_in_party + 1
			target.ai_state.ally_compassion = 10
--			target.silent_levelup = true
--			ai_state = { talent_in=1, ai_move="move_astar", ally_compassion=10 },
			target.summoner = self
			target.summoner_gain_exp = true
			target.mercenary = true
--			target.combat_armor = target.combat_armor + self:getTalentLevel(self.T_MAINTENANCE)*5
--			target.combat_dam = target.combat.dam + self:getTalentLevel(self.T_MAINTENANCE)*2
			target.inc_stats = {
				str=self:getTalentLevel(self.T_TRAINING)*2,
				dex=self:getTalentLevel(self.T_TRAINING)*2,
				mag=self:getTalentLevel(self.T_TRAINING)*2,
				wil=self:getTalentLevel(self.T_TRAINING)*2,
				cun=self:getTalentLevel(self.T_TRAINING)*2,
				con=self:getTalentLevel(self.T_TRAINING)*2,
			}

			target.on_die = function(self, who)
				game.party:removeMember(self, true)
				self:disappear()
				self:removed()
				self.summoner.mercs_in_party = self.summoner.mercs_in_party - 1
			end
			target.forceLevelup = function(self) if self.summoner then return mod.class.Actor.forceLevelup(self, self.summoner.level) end end

			if self:getTalentLevelRaw(self.T_ORDERS) == 1 then
				game.party:addMember(target, {
					control="order",
					title="Mercenary",
					orders = {target=true},
				})
			elseif self:getTalentLevelRaw(self.T_ORDERS) == 2 then
				game.party:addMember(target, {
					control="order",
					title="Mercenary",
					orders = {target=true, leash=true},
				})
			elseif self:getTalentLevelRaw(self.T_ORDERS) == 3 then
				game.party:addMember(target, {
					control="order",
					title="Mercenary",
					orders = {target=true, leash=true, anchor=true},
				})
			elseif self:getTalentLevelRaw(self.T_ORDERS) == 4 then
				game.party:addMember(target, {
					control="order",
					title="Mercenary",
					orders = {target=true, leash=true, anchor=true, talents=true},
				})
			elseif self:getTalentLevelRaw(self.T_ORDERS) >= 5 then
				game.party:addMember(target, {
					control="order",
					title="Mercenary",
					orders = {target=true, leash=true, anchor=true, talents=true, behavior=true},
				})
			else
				game.party:addMember(target, {
					control="order",
					title="Mercenary",
				})
			end
		end)
	return true
	end,

	info = function(self, t)
		local undead_addendum = function() if self.undead then return "As an undead, you can also hire other undead." else return "" end end
		local price = t.getPrice(self, t)

		return ([[By speaking quickly and waving money around, you can attempt to hire almost anyone.
		Only sentient creatures (such as humanoids, giants, orcs...) can be hired this way, and elites and bosses are immune to your influence.
		]]..undead_addendum()..[[ 
Your persuasion power is %d, and successfully hiring someone in this fashion costs %d gold.]]):format(self:combatTalentStatDamage(t, "wil", 15, 100) + self.level, price)
	end,
}

--newTalent{
--	name = "Maintenance",
--	type = {"leadership/command", 4},
--	require = lead_wil_req4,
--	points = 5,
--	mode = "passive",
--	on_learn = function(self, t)
--		for uid, e in pairs(game.level.entities) do
--			if e.mercenary and e.summoner and e.summoner == self then
--				e.combat_armor = e.combat_armor + 5
--				e.combat_dam = e.combat_dam + 2
--			end
--		end
--	end,
--	on_unlearn = function(self, t)
--		for uid, e in pairs(game.level.entities) do
--			if e.mercenary and e.summoner and e.summoner == self then
--				e.combat_armor = e.combat_armor - 5
--				e.combat_dam = e.combat_dam - 2
--			end
--		end
--	end,
--	info = function(self, t)
--		return ([[Mercenaries prefer to keep their own equipment, but they can be convinced to polish their gear once in a while.
--		Good maintenance improves the armor of your hired mercenaries by %d, and increases their damage by %d.]]):format(self:getTalentLevel(t)*5, self:getTalentLevel(t)*2)
--	end,
--}