-- ToME - Tales of Maj'Eyal
-- Copyright (C) 2009, 2010, 2011 Nicolas Casalini
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Nicolas Casalini "DarkGod"
-- darkgod@te4.org

local function getMoraleMultiplier(self, min, max)
	return (min + ((max - min) * math.min(self.morale, 10) / 10))
end

local function getDefEffect(self, t)
	return self:getWil() * self:getTalentLevel(t) * 0.05
end

local function getAtkEffect(self, t)
	return self:getWil() * self:getTalentLevel(t) * 0.05
end

local function getDamEffect(self, t)
	return self:getWil() * self:getTalentLevel(t) * 0.03
end

local function getSavEffect(self, t)
	return self:getWil() * self:getTalentLevel(t) * 0.05
end

local function getResEffect(self, t)
	return self:getWil() * self:getTalentLevel(t) * 0.04
end

newTalent{
	name = "Presence",
	type = {"leadership/presence", 1},
	mode = "sustained",
	require = lead_wil_high1,
	points = 5,
	cooldown = 1,
	sustain_stamina = 100,
	range = 2,
	tactical = { BUFF = 5 },
	activate = function(self, t)
		game:playSoundNear(self, "talents/arcane")
		return {
-- make presence particles. Blue, maybe. Also make effect particles.
			particle = self:addParticles(Particles.new("presence", 1)),
		}
	end,
	deactivate = function(self, t, p)
		self:removeParticles(p.particle)
		return true
	end,
	do_presence = function(self, tPresence)
		-- all presence effects are handled here

		if game.zone.wilderness then return end

		local tInspiration = self:getTalentFromId(self.T_INSPIRATION)
		local tIntimidation = self:getTalentFromId(self.T_INTIMIDATION)
		local tCompassion = self:getTalentFromId(self.T_COMPASSION)

		local grids = core.fov.circle_grids(self.x, self.y, self:getTalentRange(tPresence), true)
		for x, yy in pairs(grids) do
			for y, _ in pairs(grids[x]) do
				local target = game.level.map(x, y, Map.ACTOR)
				if target and self:reactionToward(target) > 0 and not (target == self) and not target:hasEffect(target.EFF_LEADERS_PRESENCE) then
					target:setEffect(target.EFF_LEADERS_PRESENCE, 3, {
					defense=getDefEffect(self, tPresence) * getMoraleMultiplier(self, 0.5, 1.5) or 0, 
					attack=getAtkEffect(self, tInspiration) * getMoraleMultiplier(self, 0.5, 1.5) or 0,
					damage=getDamEffect(self, tInspiration) * getMoraleMultiplier(self, 0.5, 1.5) or 0,
					saves=getSavEffect(self, tIntimidation) * getMoraleMultiplier(self, 0.5, 1.5) or 0,
					resist=getResEffect(self, tCompassion) * getMoraleMultiplier(self, 0.5, 1.5) or 0})
				end
			end
		end

	end,
	info = function(self, t)
		local defEffect = getDefEffect(self, t)
		return ([[Your soldiers draw strength from your presence.
		Increases the defense of nearby allies by %d (at 0.0 Morale) to %d (at 10.0 Morale).]]):format(defEffect * 0.5, defEffect * 1.5)
	end,
}

newTalent{
	name = "Inspiration",
	type = {"leadership/presence", 2},
	require = lead_wil_high2,
	points = 5,
	mode = "passive",
	info = function(self, t)
		local atkEffect = getAtkEffect(self, t)
		local damEffect = getDamEffect(self, t)

		return ([[A few simple words can bring out the best in your troops. Increases the accuracy of all allies within your presence by %d (at 0.0 Morale) to %d (at 10.0 Morale), and increases their damage by %d (at 0.0 Morale) to %d (at 10.0 Morale).]]):format(atkEffect * 0.5, atkEffect * 1.5, damEffect * 0.5, damEffect * 1.5)
	end,
}

newTalent{
	name = "Intimidation",
	type = {"leadership/presence", 3},
	require = lead_wil_high3,
	points = 5,
	mode = "passive",
	info = function(self, t)
		local savEffect = getSavEffect(self, t)

		return ([[Your soldiers refuse to show weakness when you are near. Increases the saves of all allies within your presence by %d (at 0.0 Morale) to %d (at 10.0 Morale).]]):format(savEffect * 0.5, savEffect * 1.5)
	end,
}

newTalent{
	name = "Compassion",
	type = {"leadership/presence", 4},
	require = lead_wil_high4,
	points = 5,
	mode = "passive",
	info = function(self, t)
		local resEffect = getResEffect(self, t)

		return ([[You take care of your soldiers. Allies within your presence gain %d (at 0.0 Morale) to %d (at 10.0 Morale) global damage reduction.]]):format(resEffect * 0.5, resEffect * 1.5)
	end,
}
