-- ToME - Tales of Maj'Eyal
-- Copyright (C) 2009, 2010, 2011 Nicolas Casalini
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Nicolas Casalini "DarkGod"
-- darkgod@te4.org

local function getMoraleMultiplier(self, min, max)
	return (min + ((max - min) * math.min(self.morale, 10) / 10))
end

newTalent{
	name = "Cheer",
	type = {"leadership/teamwork", 1},
	require = lead_cun_req1,
	points = 5,
	cooldown = 20,
	range = 2,
	random_ego = "attack",
	tactical = { DISABLE = 2 },

	action = function(self, t)
		for uid, e in pairs(game.level.entities) do
			local ally = game.level.map(e.x, e.y, Map.ACTOR)

			if ally and ( (ally == self) or (ally.mercenary and ally.summoner and (ally.summoner == self)) ) then
				local grids = core.fov.circle_grids(ally.x, ally.y, 2, true)
				for x, yy in pairs(grids) do
					for y, _ in pairs(grids[x]) do
--						ally:project(target, x, y, function(px, py)
						local target = game.level.map(x, y, Map.ACTOR)
							if target and self:reactionToward(target) < 0  and target:checkHit(ally:combatAttackStr(), target:combatPhysicalResist(), 0, 95, 10) then
							target:setEffect(target.EFF_BATTLE_CRY, 5, {power=6 * self:getTalentLevel(t) * getMoraleMultiplier(self, 05, 1.5)})
						end
--						end, {type="flame"})
					end
				end
			end
		end
		return true
	end,

	info = function(self, t)
		local defEffect = (6 * self:getTalentLevel(t))

		return ([[Your troops shout in unison, breaking the spirits of your foes. All foes within a radius of 2 of each of your units (including yourself) have their defense lowered by %d (at 0.0 Morale) to %d (at 10.0 Morale).]]):format(defEffect * 0.5, defEffect * 1.5)
	end,
}

newTalent{
	name = "Formation",
	type = {"leadership/teamwork", 2},
	require = lead_cun_req2,
	points = 5,
	cooldown = 20,
--	morale = -0.3,
	range = 1,
	tactical = { BUFF = 2 },

	action = function(self, t)
		for uid, e in pairs(game.level.entities) do
			local ally = game.level.map(e.x, e.y, Map.ACTOR)
	
			if ally and ( (ally == self) or (ally.mercenary and ally.summoner and (ally.summoner == self)) ) then
				local grids = core.fov.circle_grids(ally.x, ally.y, 1, true)
				for x, yy in pairs(grids) do
					for y, _ in pairs(grids[x]) do
						local target = game.level.map(x, y, Map.ACTOR)
						if target and self:reactionToward(target) > 0 and not (target == ally) then
							target:setEffect(target.EFF_FORMATION, 10, {
								defense = 6 * self:getTalentLevel(t) * getMoraleMultiplier(self, 0.5, 1.5),
								armor = 6 * self:getTalentLevel(t) * getMoraleMultiplier(self, 0.5, 1.5)
							})
						end
					end
				end
			end
		end
		return true
	end,
 
	info = function(self, t)
		local defEffect = (6 * self:getTalentLevel(t))
		local armEffect = (6 * self:getTalentLevel(t))

		return ([[Your men guard each other using rehearsed battle tactics. If two or more of your units are adjacent to each other, both gain %d (at 0.0 Morale) to %d (at 10.0 Morale) defense, and %d (at 0.0 Morale) to %d (at 10.0 Morale) armor.
		The effect lasts for 10 turns.]]):format(defEffect * 0.5, defEffect * 1.5, armEffect * 0.5, armEffect * 1.5)
	end,
}

newTalent{
	name = "Report",
	type = {"leadership/teamwork", 3},
	require = lead_cun_req3,
	points = 5,
	cooldown = 30,
	no_npc_use = true,
	tactical = { BUFF = 2 },
	getRadius = function(self, t) return self:getTalentLevel(self.T_REPORT) end,
	action = function(self, t)
		self:setEffect(self.EFF_REPORT, 5 + self:getTalentLevel(t) * 2, {radius=t.getRadius(self, t)})

--		for uid, e in pairs(game.level.entities) do
--			local target = game.level.map(e.x, e.y, Map.ACTOR)
--			if target and target.mercenary and target.summoner and (target.summoner == self) then
--				target:setEffect(target.EFF_REPORTING, self:getTalentLevel(t) * 2, {radius=self:getTalentLevel(t)})
--			end
--		end
		return true
	end,
	info = function(self, t)
	
		return([[Call for a report of the situation. Allows you to see in a radius of %d around each of your units.
		This effect will last for %d turns.]]):format(self:getTalentLevel(t), 5 + self:getTalentLevel(t) * 2)
	end,
}

newTalent{
	name = "Swarm",
	type = {"leadership/teamwork", 4},
	require = lead_cun_req4,
	points = 5,
	cooldown = 35,
	action = function(self, t)
		for uid, e in pairs(game.level.entities) do
			local target = game.level.map(e.x, e.y, Map.ACTOR)

			if target and self:reactionToward(target) < 0 then
				local count = 0
				local grids = core.fov.circle_grids(target.x, target.y, 1, true)
				for x, yy in pairs(grids) do
					for y, _ in pairs(grids[x]) do
						local ally = game.level.map(x, y, Map.ACTOR)
						if ally and ( (ally == self) or (ally.mercenary and ally.summoner and (ally.summoner == self)) ) then
							count = count +1
						end
					end
				end
				if (count >= 3) or ( (count >= 2) and (self:getTalentLevel(t) >= 5) ) then
					for x, yy in pairs(grids) do
						for y, _ in pairs(grids[x]) do
							local ally = game.level.map(x, y, Map.ACTOR)
							if ally and ( (ally == self) or (ally.mercenary and ally.summoner and (ally.summoner == self)) ) then
								local multiplier = ally:combatTalentWeaponDamage(t, 0.7, 1.9) * getMoraleMultiplier(self, 0.3, 1.0)
								ally:attackTarget(target, nil, multiplier, true)
							end
						end
					end
				end
			end
		end
		return true
	end,
	info = function(self, t)
		local multiplier = self:combatTalentWeaponDamage(t, 0.7, 1.9)
	
		return([[If three or more of your units are adjacent to an opponent, all three will make a free attack at that opponent.
		This attack is made at %d%% damage (at 0.0 morale) to %d%% damage (at 10.0 morale).
		At talent level 5 this works with with two or more units adjacent to the opponent.]]):format(multiplier * 30, multiplier * 100)
	end,
}