-- ToME - Tales of Maj'Eyal
-- Copyright (C) 2009, 2010, 2011 Nicolas Casalini
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Nicolas Casalini "DarkGod"
-- darkgod@te4.org

newTalent{
	name = "First Aid",
	type = {"technique/support", 1},
	require = techs_cun_req1,
	random_ego = "defensive",
	points = 5,
	stamina = 10,
	cooldown = 6,
	range = 1,
	requires_target = true,
	no_npc_use = true,
	action = function(self, t)
		local tg = {default_target=self, type="hit", nowarning=true, range=self:getTalentRange(t), first_target="friend"}
		local x, y, target = self:getTarget(tg)
		if not x or not y or not target then return nil end
		if math.floor(core.fov.distance(self.x, self.y, x, y)) > 1 then return nil end
		target:heal(10 + self:combatTalentStatDamage(t, "cun", 10, 200))
		target:incAir(100)
		game:playSoundNear(self, "talents/heal")
		return true
	end,
	info = function(self, t)
		return ([[Quickly bandage a target's wounds, healing it for %d.
		If the target is suffocating, this will also restore its breath.
		Heal will increase with your Cunning stat.]]):
		format(10 + self:combatTalentStatDamage(t, "cun", 10, 200))
	end,
}

newTalent{
	name = "Call",
	type = {"technique/support", 2},
	require = techs_cun_req2,
	points = 5,
	random_ego = "defensive",
	cooldown = function(self, t) return 20 - (2*math.floor(self:getTalentLevel(t))) end,
	stamina = 20,
	tactical = { CLOSEIN = 2 },
	range = function(self, t) return 2 + math.floor(self:getTalentLevel(t)) end,
	requires_target = true,

	action = function(self, t)
		local tg = {type="hit", range=self:getTalentRange(t), talent=t, first_target="friend"}
		local x, y, target = self:getTarget(tg)

		if not x or not y or not target or self:reactionToward(target) < 0 then return nil end
		if math.floor(core.fov.distance(self.x, self.y, x, y)) > self:getTalentRange(t) then return nil end

		local tx, ty = util.findFreeGrid(self.x, self.y, 5, true, {[Map.ACTOR]=true})
		if tx and ty and target:canBe("teleport") then
			target:move(tx, ty, true)
			game.logSeen(target, "%s, return!", target.name:capitalize())
		end
		return true
	end,

	info = function(self, t)
		return ([[Call a nearby ally to your side.]])
	end,
}

newTalent{
	name = "Medicine",
	type = {"technique/support", 3},
	require = techs_cun_req3,
	points = 5,
	random_ego = "defensive",
	stamina = 20,
	cooldown = 15,
	tactical = { HEAL = 2.5 },
	range = 1,
	requires_target = true,
	no_npc_use = true,
	getRegeneration = function(self, t) return self:combatTalentStatDamage(t, "cun", 10, 50) end,
	getDuration = function(self, t) return 2 + math.ceil(self:getTalentLevel(t)) end,
	getRemoveCount = function(self, t) return math.floor(self:getTalentLevel(t)) end,

	action = function(self, t)
		local tg = {default_target=self, type="hit", nowarning=true, range=self:getTalentRange(t), first_target="friend"}
		local x, y, target = self:getTarget(tg)
		if not x or not y or not target then return nil end
		if math.floor(core.fov.distance(self.x, self.y, x, y)) > 1 then return nil end

		target:setEffect(target.EFF_REGENERATION, t.getDuration(self, t), {power=t.getRegeneration(self, t)})
		
		local effs = {}
		-- Go through all spell effects
		for eff_id, p in pairs(target.tmp) do
			local e = target.tempeffect_def[eff_id]
			if e.status == "detrimental" then
				effs[#effs+1] = {"effect", eff_id}
			end
		end

		for i = 1, t.getRemoveCount(self, t) do
			if #effs == 0 then break end
			local eff = rng.tableRemove(effs)

			if eff[1] == "effect" then
				target:removeEffect(eff[2])
			end
		end
		game:playSoundNear(self, "talents/heal")
		return true
	end,

	info = function(self, t)
		local regen = t.getRegeneration(self, t)
		local duration = t.getDuration(self, t)
		local count = t.getRemoveCount(self, t)

		return ([[Cure and restore an ally, removing up to %d negative effects from your target and healing them %d health each turn for %d turns.
		The life healed will increase with the Cunning stat.]]):
		format(count, regen, duration)
	end,
}

newTalent{
--Alternative: Make an ability which reduces cooldown on target's abilities.
	name = "Assist",
	type = {"technique/support", 4},
	require = techs_cun_req4,
	points = 5,
--	random_ego = "defensive",
	stamina = 30,
	cooldown = 30,
	range = 1,
	getPower = function(self, t) return self:getTalentLevel(t)/3 end,
	getCrit = function(self, t) return self:getTalentLevel(t) * 10 end,
	
	action = function(self, t)
		local tg = {default_target=self, type="hit", nowarning=true, range=self:getTalentRange(t), first_target="friend"}
		local x, y, target = self:getTarget(tg)
		if not x or not y or not target or (target == self) then return nil end
		if math.floor(core.fov.distance(self.x, self.y, x, y)) > 1 then return nil end
		target:setEffect(target.EFF_ASSISTED, 3, {power=t.getPower(self, t), crit=t.getCrit(self, t)})
		return true
	end,
	info = function(self, t)
		local power = t.getPower(self, t)
		local crit = t.getCrit(self, t)
		return ([[Set up an ally, boosting its speed by %d%% and its crit rate by %d%% for 3 turns.]]):format(power * 100, crit)
	end,
}