-- ToME - Tales of Maj'Eyal
-- Copyright (C) 2009, 2010, 2011 Nicolas Casalini
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Nicolas Casalini "DarkGod"
-- darkgod@te4.org

local Talents = require("engine.interface.ActorTalents")

newEntity{ define_as = "CITLALI",
	type = "humanoid", subtype = "cornac",
	display = "@",
	resolvers.generic(function(e)
		image = "player/cornac_female.png"
		e.moddable_tile = "human_female"
		e.moddable_tile_base = "base_cornac_01.png"
	end),
	faction = "sunwall",
	name = "Citlali", color=colors.CRIMSON, unique = true,
	desc = [[A lean woman with a wide brimmed hat and a cloak slit for horseback riding. She carries a pair of wands in her belt.]],
	female = true,
	level_range = {30, nil}, exp_worth = 2,
	rank = 4,
	size_category = 3,
	max_life = 500, life_rating = 16, fixed_rating = true,
	infravision = 10,
	stats = { dex=40, cun=40, mag=50, },
	instakill_immune = 1,
	teleport_immune = 1,
	move_others=true,
	anger_emote = "Eastern justice!",
	on_die = function(self, who)
		game.zone.decay.respawn = false
	end,
	mana_regen = 10,
	positive_regen = 5,
	negative_regen = 5,
	open_door = true,
	autolevel = "roguemage",
	ai = "tactical", ai_state = { talent_in=1, ai_move="move_astar", },
	ai_tactic = resolvers.tactic"ranged",
	resolvers.inscriptions(3, "rune"),
	body = { INVEN = 10, MAINHAND=1, OFFHAND=1, BODY=1, BELT=1, FEET=1, HEAD=1, CLOAK=1,},
	resolvers.drops{chance=100, nb=5, {tome_drops="boss"} },
	resolvers.equip{
--		{type="weapon", subtype="staff", autoreq=true},
		{type="armor", subtype="head", autoreq=true},
		{type="armor", subtype="light", autoreq=true},
		{type="armor", subtype="feet", autoreq=true},
		{type="armor", subtype="belt", autoreq=true},
	},
	resolvers.talents{
		[Talents.T_EARTHEN_MISSILES]=8,
		[Talents.T_CHANT_OF_RESISTANCE]=6,
		[Talents.T_CIRCLE_OF_BLAZING_LIGHT]=8,
		[Talents.T_BLOOD_RED_MOON]=8,
		[Talents.T_TOTALITY]=3,
		[Talents.T_CORONA]=6,
		[Talents.T_HYMN_OF_DETECTION]=8,
		[Talents.T_BARRIER]=5,
		[Talents.T_PROVIDENCE]=8,
		[Talents.T_MOONLIGHT_RAY]=5,
		[Talents.T_TWILIGHT_SURGE]=5,
		[Talents.T_STARFALL]=6,
		[Talents.T_FIREBEAM]=6,
	},
	resolvers.sustains_at_birth(),
	can_talk = "citlali",
}

newEntity{
	define_as = "BASE_NPC_TAVERN",
	type = "humanoid", subtype = "cornac",
	display = "@", color=colors.WHITE,
	resolvers.generic(function(e)
		if rng.percent(50) then
			e.female = true
			image = "player/cornac_female.png"
		else
			image = "player/cornac_male.png"
		end
		e.moddable_tile = "human_female"
		e.moddable_tile_base = "base_cornac_01.png"
	end),

	faction = "sunwall",
	anger_emote = "Its gone mad!",
	resolvers.racial(),
	combat = { dam=resolvers.rngavg(1,2), atk=2, apr=0, dammod={str=0.4} },
	body = { INVEN = 10, MAINHAND=1, OFFHAND=1, BODY=1, QUIVER=1 },
	lite = 3,
	life_rating = 10,
	rank = 2,
	size_category = 3,
	positive_regen = 5,
	negative_regen = 5,
	ai = "tactical", ai_state = { ai_move="move_dmap", talent_in=3, },
--	emote_random = resolvers.emote_random{allow_backup_guardian=true},
	can_talk = "mercenary",
}

newEntity{ base = "BASE_NPC_TAVERN",
	name = "Paladin", color=colors.LIGHT_RED,
	desc = [[A warrior who channels the power of the sun.]],
	level_range = {1, nil}, exp_worth = 1,
	rarity = 1,
	max_life = resolvers.rngavg(90,100),
	stats = { wil=15, cun=14 },
	autolevel = "warriormage",
	resolvers.equip{
		{type="weapon", subtype="longsword", autoreq=true},
		{type="armor", subtype="heavy", autoreq=true},
		{type="armor", subtype="shield", autoreq=true},
	},
	combat_armor = 5, combat_def = 0,
	resolvers.talents{
		[Talents.T_ARMOUR_TRAINING]={base=3, every=4, max=8},
		[Talents.T_CHANT_OF_FORTITUDE]={base=1, every=5, max=8},
--		[Talents.T_WEAPON_OF_LIGHT]={base=1, every=5, max=8},
		[Talents.T_WAVE_OF_POWER]={base=1, every=5, max=8},
--		[Talents.T_CRUSADE]={base=1, every=5, max=8},
--		[Talents.T_SHIELD_OF_LIGHT]={base=1, every=5, max=8},
--		[Talents.T_SECOND_LIFE]={base=1, every=5, max=8},
		[Talents.T_HEALING_LIGHT]={base=1, every=5, max=8},
		[Talents.T_PROVIDENCE]={base=1, every=5, max=8},
		[Talents.T_BATHE_IN_LIGHT]={base=1, every=5, max=8},
--		[Talents.T_FIREBEAM]={base=1, every=5, max=8},
	},
	max_inscriptions = 3,
--	resolvers.inscriptions(1, "infusion"),
}

newEntity{ base = "BASE_NPC_TAVERN",
	name = "Anorithil", color=colors.LIGHT_RED,
	desc = [[One who walks the line between light and darkness.]],
	level_range = {1, nil}, exp_worth = 1,
	rarity = 1,
	max_life = resolvers.rngavg(80,90),
	stats = { mag=15, cun=14 },
	autolevel = "roguemage",
	resolvers.equip{
		{type="weapon", subtype="staff", autoreq=true},
		{type="armor", subtype="cloth", autoreq=true},
	},
	combat_armor = 0, combat_def = 0,
	resolvers.talents{
--		[Talents.T_SUNBURST]={base=1, every=5, max=8},
--		[Talents.T_BLOOD_RED_MOON]={base=1, every=5, max=8},
		[Talents.T_BARRIER]={base=1, every=5, max=8},
		[Talents.T_PROVIDENCE]={base=1, every=5, max=8},
		[Talents.T_MOONLIGHT_RAY]={base=1, every=5, max=8},
--		[Talents.T_TWILIGHT_SURGE]={base=1, every=5, max=8},
--		[Talents.T_CORONA]={base=1, every=5, max=8},
		[Talents.T_HYMN_OF_PERSEVERANCE]={base=1, every=5, max=8},
		[Talents.T_CHANT_OF_LIGHT]={base=1, every=5, max=8},
		[Talents.T_FIREBEAM]={base=1, every=5, max=8},
	},
	max_inscriptions = 3,
--	resolvers.inscriptions(1, "infusion"),
}