-- ToME - Tales of Maj'Eyal
-- Copyright (C) 2009, 2010, 2011 Nicolas Casalini
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Nicolas Casalini "DarkGod"
-- darkgod@te4.org

local Talents = require("engine.interface.ActorTalents")

newEntity{ define_as = "WESTREACH",
	type = "humanoid", subtype = "human",
	display = "p",
	resolvers.generic(function(e)
		image = "player/cornac_male.png"
		e.moddable_tile = "human_male"
		e.moddable_tile_base = "base_cornac_01.png"
	end),
	faction = "allied-kingdoms",
	name = "Westreach", color=colors.CRIMSON, unique = true,
	desc = [[A large man in a leather apron.]],
	level_range = {30, nil}, exp_worth = 2,
	rank = 4,
	size_category = 3,
	stamina_regen = 120,
	max_stamina = 2000,
	max_life = 350, life_rating = 24, fixed_rating = true,
	infravision = 10,
	stats = { str=45, dex=40, cun=24, mag=10, con=42 },
	instakill_immune = 1,
	teleport_immune = 1,
	move_others=true,
	anger_emote = "No fighting in the bar!",
	on_die = function(self, who)
		game.zone.decay.respawn = false
	end,
	open_door = true,
	autolevel = "archer",
	ai = "tactical", ai_state = { talent_in=1, ai_move="move_astar", },
	ai_tactic = resolvers.tactic"ranged",
	resolvers.inscriptions(3, "rune"),
	resolvers.inscriptions(1, {"manasurge rune"}),
	body = { INVEN = 10, MAINHAND=1, OFFHAND=1, BODY=1 },
	resolvers.drops{chance=100, nb=5, {tome_drops="boss"} },
	resolvers.equip{
		{type="weapon", subtype="bow", autoreq=true, tome_drops="store"},
		{type="armor", subtype="light", autoreq=true, tome_drops="store"},
	},
	resolvers.talents{
		[Talents.T_SHOOT] = 1,
		[Talents.T_BOW_MASTERY] = 6,
		[Talents.T_RAPID_SHOT] = 5,
		[Talents.T_STEADY_SHOT] = 5,
		[Talents.T_FLARE] = 3,
		[Talents.T_PINNING_SHOT] = 4,
	},
	resolvers.sustains_at_birth(),
	can_talk = "westreach",
}

newEntity{
	define_as = "BASE_NPC_TAVERN",
	type = "humanoid", subtype = "cornac",
	display = "p", color=colors.WHITE,
	resolvers.generic(function(e)
		if rng.percent(50) then
			e.female = true
			image = "player/cornac_female.png"
		else
			image = "player/cornac_male.png"
		end
		e.moddable_tile = "human_#sex#"
		e.moddable_tile_base = "base_cornac_01.png"
	end),

	faction = "allied-kingdoms",
	anger_emote = "Catch @himher@!",
	resolvers.racial(),
	combat = { dam=resolvers.rngavg(1,2), atk=2, apr=0, dammod={str=0.4} },
	body = { INVEN = 10, MAINHAND=1, OFFHAND=1, BODY=1, QUIVER=1 },
	lite = 3,
	life_rating = 11,
	rank = 2,
	size_category = 3,
	ai = "tactical", ai_state = { ai_move="move_dmap", talent_in=3, },
	stats = { str=10, dex=10, mag=10, wil=10, con=10 },
--	emote_random = resolvers.emote_random{allow_backup_guardian=true},
	can_talk = "mercenary",	

}

newEntity{ base = "BASE_NPC_TAVERN",
	name = "Fighter", color=colors.LIGHT_RED,
	desc = [[An armored warrior.]],
	level_range = {1, nil}, exp_worth = 1,
	rarity = 3,
	max_life = resolvers.rngavg(70,80),
	stats = { str=14, dex=13, con=12 },
	autolevel = "warrior",
	resolvers.equip{
		{type="weapon", subtype="longsword", autoreq=true},
		{type="armor", subtype="shield", autoreq=true},
		{type="armor", subtype="heavy", autoreq=true},
	},
	combat_armor = 5, combat_def = 0,
	resolvers.talents{
		[Talents.T_ARMOUR_TRAINING]={base=3, every=4, max=10},
		[Talents.T_BLEEDING_EDGE]={base=1, every=5, max=8},
		[Talents.T_SHIELD_PUMMEL]={base=1, every=5, max=8},
		[Talents.T_JUGGERNAUT]={base=1, every=5, max=8},
		[Talents.T_OVERPOWER]={base=1, every=5, max=8},
		[Talents.T_SHIELD_EXPERTISE]={base=1, every=5, max=8},
	},
	max_inscriptions = 3,
--	resolvers.inscriptions(1, "infusion"),
	resolvers.inscriptions(1, {"regeneration infusion"}),
}

newEntity{ base = "BASE_NPC_TAVERN",
	name = "Warrior", color=colors.LIGHT_RED,
	desc = [[A heavily muscled warrior wielding a two handed sword.]],
	level_range = {1, nil}, exp_worth = 1,
	rarity = 3,
	max_life = resolvers.rngavg(90,100),
	stats = { str=18, con=12 },
	autolevel = "zerker",
	resolvers.equip{
		{type="weapon", subtype="greatsword", autoreq=true},
		{type="armor", subtype="light", autoreq=true},
	},
	combat_armor = 5, combat_def = 0,
	resolvers.talents{
		[Talents.T_ARMOUR_TRAINING]=3,
--		[Talents.T_DEATH_BLOW]={base=1, every=5, max=8},
		[Talents.T_WARSHOUT]={base=1, every=5, max=8},
		[Talents.T_SECOND_WIND]={base=1, every=5, max=8},
		[Talents.T_GREATER_WEAPON_FOCUS]={base=1, every=5, max=8},
		[Talents.T_BLOODBATH]={base=1, every=5, max=8},
		[Talents.T_BLOODRAGE]={base=1, every=5, max=5},
	},
	max_inscriptions = 3,
--	resolvers.inscriptions(1, "infusion"),
	resolvers.inscriptions(1, {"regeneration infusion"}),
}
newEntity{ base = "BASE_NPC_TAVERN",
	name = "Archer", color=colors.UMBER,
	desc = [[An agile hunter with a bow slung over its shoulder.]],
	level_range = {1, nil}, exp_worth = 1,
	rarity = 2,
	max_life = resolvers.rngavg(70,80),
	autolevel = "archer",
	stats = { str=14, dex=15},
	resolvers.equip{
		{type="weapon", subtype="longbow", autoreq=true},
		{type="armor", subtype="light", autoreq=true},
	},
	combat_armor = 5, combat_def = 0,
	resolvers.talents{
		[Talents.T_ARMOUR_TRAINING]=1,
		[Talents.T_SHOOT] = 1,
		[Talents.T_BOW_MASTERY] = {base=1, every=5, max=8},
		[Talents.T_STEADY_SHOT] = {base=1, every=5, max=5},
		[Talents.T_PINNING_SHOT] = {base=1, every=5, max=5},
	},
	max_inscriptions = 3,
	resolvers.inscriptions(1, {"regeneration infusion"}),
}
